# -*- coding: utf-8 -*-
# Author: Marcin Kasperski

"""
Converting SuperMemo "text" export to Mnemosyne XML format.
I used this script to successfully import Advanced English SpeedUp!
and Proficiency with sound (detailed article in Polish - see
http://notatnik.mekk.waw.pl/archives/125-SuperMemo_pod_Linuksem.html )

SuperMemo "text" format handled here:

Begin Element #1
  Source=Y:\SuperMemo\AdvEngSpeedUp\systems\SPEEDUP!
  Parent=0
  ParentTitle=
  Begin ElementInfo #1
    Title=Speed Up!
    Type=Topic
    Status=Dismissed
    FirstGrade=6
    (...)

Mnemosyne XML: http://www.mnemosyne-proj.org/help/mnemosyne-xml-dtd.php

Warning: the script does not copy learning data. It would not be difficult
to add (search for "TODO: copy learning data" below) but one would need
to know how to map SuperMemo statistics onto Mnemosyne ones.

This module can be used as script or as module. When used as script
it should be run as:

    python smexport_to_mnemosyne_xml.py config.ini

where config.ini looks like:

    CONVERT = dict(

      # File exported from SuperMemo:
      input_file  = "/home/mekk/SpeedUpExport.txt",

      # Output file:
      output_file = "/home/mekk/SpeedUpExport.xml",

      # Function called to translate original sound paths
      # to destination paths. Spelled as below it translates
      #   'E:\SM7\systems\AdvEnglish\elements\1\33\24.wav'
      # to
      #   'advenglish/sounds/1/33/24.ogg'
      name_mapper = FileNameMapper(
        orig_root = 'Y:\AdvEngSpeedUp\systems\SPEEDUP!\elements', 
        dest_root = "SuperMemo/SpeedUpElements",
        convert_slashes = True,
        convert_ext = { 'wav' : 'ogg' }),

      # Function called to generate mnemosyne category from SuperMemo
      # hierarchical category. Spelled as below it translates
      #
      #    ('Speed Up!', 'Advanced English', 'General English', 'Body Life and Living Creatures', "Head and the Face', '#1 Head and the Face')
      #
      # into
      # 
      #    'Advanced_English-Body_Life_and_Living_Creatures-Head_and_the_Face'
      #
      category_mapper = CategoryMapper(
        re.compile('^(Speed Up!|Advanced English|\[\d.*|\#\d.*)$')),
    )

Created XML can be imported to Mnemosyne using XML importer. In case
invalid characters are reported, fix them manually (or improve
escape_xml routine below and rerun).

To handle sound properly, copy/mount and if necessary convert the sound
files. In example above I copied 'elements' directory as 
~/.mnemosyne/SuperMemo/SpeedUpElements and converted all .wav files
to .ogg using "sox -t mp3 file.wav file.ogg" for every file.

"""

class FileNameMapper(object):
    """
    Converting paths to sound and image elements. This class is able to
    change from
        E:\SM7\systems\AdvEnglish\elements\1\33\24.wav
    to
        advenglish/sounds/1/33/24.ogg
    (for example). For this case it should be callled with
        FileNameMapper(r'E:\SM7\systems\AdvEnglish\elements',
                       'advenglish/sounds',
                       True,
                       { 'wav' : 'ogg' })
    """
    def __init__(self, orig_root, dest_root, convert_slashes, convert_ext):
        self.orig_root = orig_root
        self.dest_root = dest_root
        self.convert_slashes = convert_slashes
        self.convert_ext = convert_ext
    def __call__(self, filename):
        rest = filename.replace(self.orig_root, "")
        if self.convert_slashes:
            rest = rest.replace("\\", "/")
        if self.convert_ext:
            for sext, dext in self.convert_ext.iteritems():
                rest = rest.replace("." + sext, "." + dext)
        return self.dest_root + rest

class CategoryMapper(object):
    """
    Constructing destination category from supermemo category path.
    This class converts sth like

    [ 'Speed Up!', 'Advanced English', 'General English', 'Body Life and Living Creatures', "Head and the Face', '#1 Head and the Face']

    into

    'Advanced_English-Body_Life_and_Living_Creatures-Head_and_the_Face'

    For this purpose:
    
    CategoryMapper(re.compile('^(Speed Up!|General English|\#.*)$'))
    """
    def __init__(self, skip_re):
        self.skip_re = skip_re
    def __call__(self, *catpath):
        return "-".join([ escape_category(cat) for cat in catpath if not self.skip_re.match(cat)])



import pyparsing as pyp
from xml.sax.saxutils import escape as orig_escape_xml
import re

def escape_xml(txt):
    # Hack around invalid characters found in SuperMemo export
    txt = txt.replace('', 'è')
    txt = txt.replace('', 'ñ')
    txt = txt.replace('', 'Ä')
    txt = txt.replace('', 'è')
    txt = txt.replace('', 'è') # Can be wrong, sth with grave accent
    txt = txt.replace('', 'ó') # Can be wrong, sth with acute accent
    txt = txt.replace('', '°')
    txt = txt.replace('', '·')
    txt = txt.replace('', 'ê')
    txt = txt.replace('', 'ó')
    return orig_escape_xml(txt)

re_notnm = re.compile("([^-_a-zA-Z0-9]+)")
def escape_category(txt):
    """
    Stronger than escape_xml as it must provide tokens (and for example
    'Speed Up!' fails to load)
    """
    return re_notnm.sub("_", txt)

class Converter(object):

    digits = "0123456789"
    hash_id = pyp.Literal("#").suppress() + pyp.Word(digits).setResultsName("id")
    equals = pyp.Literal("=").suppress()

    # replace pyp.restOfLine to avoid eating \r
    restOfLine = pyp.Regex(r"[^\r\n]*").leaveWhitespace()

    setting = pyp.Dict(pyp.Group(pyp.Word(pyp.alphas) + pyp.Literal("=").suppress() + restOfLine))
    #setting_list = pyp.OneOrMore(setting)
    setting_list = pyp.dictOf(
        pyp.Word(pyp.alphas) + equals, restOfLine)

    element_info =  pyp.Group(
        pyp.Literal("Begin ElementInfo") + hash_id \
        + setting_list \
        + pyp.Suppress(pyp.Literal("End ElementInfo") + hash_id)
        ).setResultsName("info")

    component =  pyp.Group(
        pyp.Literal("Begin Component").suppress() + hash_id \
        + setting_list \
        + pyp.Suppress(pyp.Literal("End Component") + hash_id)
        ).setResultsName("component", True)

    element = pyp.Literal("Begin Element").suppress() + hash_id \
        + pyp.OneOrMore( setting | component | element_info ) \
        + pyp.Suppress(pyp.Literal("End Element") + hash_id)

    grammar = pyp.OneOrMore(element)

    def convert(self, input_file, output_file, 
                name_mapper = None, category_mapper = None):
        f = open(input_file, "rb")
        file_contents = f.read()
        f.close()

        ### Informacje o kategoriach
        
        # id dziecka -> id rodzica
        cat_id2parentid = {}
        # id -> orginalne name
        cat_id2name = {}
        # id -> ilość item
        cat_id2count = {}
        # id -> wynikowe name
        cat_id2final = {}

        categories = []  # Identyfikatory kategorii w kolejności występowania
        items = []       # Gotowe XMLe

        for tokens, startloc, endloc in self.element.scanString(file_contents):

            if tokens.info.Type == "Topic":
                cat_id2parentid[tokens.id] = tokens.Parent
                cat_id2name[tokens.id] = tokens.info.Title
                cat_id2count[tokens.id] = 0

                catpath = [ tokens.info.Title ]
                parid = tokens.Parent
                while parid and parid != '0':
                    catpath.append(cat_id2name[parid])
                    parid = cat_id2parentid.get(parid)
                catpath.reverse()
                if category_mapper:
                    final = category_mapper(*catpath)
                else:
                    final = escape_category("-".join(catpath))
                cat_id2final[tokens.id] = final
                cat_id2count[tokens.id] = 0

                categories.append(tokens.id)

                print "Category %s (mapped as %s)" % (tokens.info.Title, final)

            elif tokens.info.Type == "Item":

                #print "Item", tokens.info.Title, "in", tokens.ParentTitle

                t = "<item>\n"

                # TODO: copy learning data to appropriate <item> attributes
                # Those are available as tokens.info.Status, 
                # tokens.info.Ordinal, tokens.info.ForgettingIndex etc
                #
                # Full list of items and example values:
                #         Status Pending
                #         Ordinal 1208.000000
                #         ForgettingIndex 10
                #         Repetitions 0
                #         Interval 0
                #         FirstGrade 6
                #         AFactor 3.000
                #         Lapses 0
                #         UFactor 0.000
                #         RepetitionHistory 0
                #         LastRepetition 0
                #
                # Mnemosyne notation: 
                # <item gr="2" e="2.500" ac_rp="2" rt_rp="0" lps="0"
                #       ac_rp_l="2" rt_rp_l="0" l_rp="32" n_rp="34">
                # where
                #   gr= the grade, default is 0
                #   e= easiness
                #   ac_rp= acquisition reps
                #   rt_rp= retention reps
                #   lps= lapses
                #   ac_rp_l, rt_rp_l = ... since lapse
                #   l_rp= last rep
                #   n_rp= next rep

                # Uncomment to dump above to the standard output
                #for k in tokens.info.keys():
                #    if not k in ['Title', 'Type']:
                #        print "        ", k, tokens.info[k]

                cat = cat_id2final[tokens.Parent]
                cat_id2count[tokens.Parent] += 1
                t += "  <cat>&lt;%s&gt;</cat>\n" % cat
                #escape_category(tokens.ParentTitle)

                fields = dict(Q = [], A = [])
                curr = "Q"

                for c in tokens.component:
                    #print "  Component", c.Type, c.id

                    if c.FontMember == "Answer": 
                        # Switch to answering
                        curr = "A"
                    
                    if c.Type == "Text":
                        fields[curr].append(c.Text.replace("", "\n"))
                    elif c.Type == "Sound":
                        name = c.SoundFile
                        if name_mapper:
                            name = name_mapper(name)
                        fields[curr].append(
                            '<sound src="%s">' % name)
                    elif c.Type == "Image":
                        name = c.ImageName
                        if name_mapper:
                            name = name_mapper(name)
                        fields[curr].append(
                            '<img src="%s">' % name)
                    elif c.Type == "RTF":
                        pass
                    else:
                        raise Exception("Unknown component type: '%s'" % c.Type)

                    # Uncomment to dump data
                    #for l in c.keys():
                        #print "     ", l, c[l]

                t += " <Q>%s</Q>\n" % escape_xml("\n".join(fields["Q"]))
                t += " <A>%s</A>\n" % escape_xml("\n".join(fields["A"]))
                t += "</item>\n"

                items.append(t)
                #print items[-1]
            else:
                raise Exception("Unknown element type: '%s'" % tokens.info.Type)

        g = open(output_file, "w")
        g.write("""<?xml version="1.0" encoding="UTF-8"?>
<mnemosyne core_version="1">
""")

        written_categories = {}
        # Duplicate skipping, we can map many initial categories to one resulting
        # (Health/#1 Health and Health/#2 Health can be both mapped as Health)
        for cid in categories:
            if cat_id2count[cid] > 0:
                cname = cat_id2final[cid]
                if not cname in written_categories: 
                    written_categories[cname] = 1
                    g.write("""<category active="1">
<name>&lt;%s&gt;</name>
</category>
""" % cname)

        g.write("\n".join(items))
        g.write("</mnemosyne>\n")
        g.close()

if __name__ == "__main__":
    import sys
    if len(sys.argv) != 2:
        print """Usage:
    %s config.ini
""" % sys.argv[0]
        sys.exit(1)

    text = file(sys.argv[1]).read()
    exec(text, globals(), locals())

    if CONVERT:
        print "Converting %s to %s" % (CONVERT['input_file'], 
                                       CONVERT['output_file'])
        conv = Converter()
        conv.convert(**CONVERT)
        print "Done"
    else:
        print "INI file should define CONVERT variable"
